package de.unisaar.jfap.ex8;

import de.unisaar.jfap.ex7.DirectedGraph;
import de.unisaar.jfap.ex7.Edge;
import java.util.Comparator;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.PriorityQueue;

/**
 * Calculate single source shortest paths in a graph using Dijkstra's algorithm.
 *
 * @author Bernd Kiefer, DFKI
 * @author Jörg Steffen, DFKI
 */
public class DijkstraShortestPath {

  private DirectedGraph<Integer, ?> graph;

  // store for each vertex its distance to start vertex
  private Map<Integer, Integer> distances;

  /**
   * Create a new instance to calculate a shortest path in the given directed graph.
   *
   * @param graph a directed graph with integers as edge weights
   */
  public DijkstraShortestPath(DirectedGraph<Integer, ?> graph) {
    this.graph = graph;
    this.distances = new HashMap<>();
  }

  /** @return the distances map resulting from the last {@link #shortestPath(int, int)} call. */
  public Map<Integer, Integer> getDistances() {
    return this.distances;
  }

  /**
   * Calculate the shortest path from the given start vertex to the given end vertex.
   *
   * @param startVertex the start vertex
   * @param endVertex the end vertex
   * @return an edges list representing the shortest path or {@code null} if there is no path
   *     between the start vertex and the end vertex
   */
  public List<Edge<Integer>> shortestPath(int startVertex, int endVertex) {

    this.distances.clear();

    // handle special case of identical start and end vertex
    if (startVertex == endVertex) {
      return new LinkedList<Edge<Integer>>();
    }

    // store for each vertex the edge starting at the predecessor vertex if existing
    Map<Integer, Edge<Integer>> predecessor = new HashMap<>();

    // init comparator to be used in priority queue
    Comparator<Integer> vertComp = new DistComp(this.distances);

    // init priority queue returning the vertex with the lowest distance
    PriorityQueue<Integer> queue =
        new PriorityQueue<>((this.graph.getNumberOfVertices() / 4) + 1, vertComp);

    this.distances.put(startVertex, 0);
    // we will use the fact that a vertex is not in distances to indicate its
    // value is +infinity; see compare method in DistComp

    // add all vertices to the queue
    for (int vertex : graph.getVertices()) {
      queue.offer(vertex);
    }

    while (!queue.isEmpty()) {
      int minWeightVertex = queue.poll();
      // stop if end vertex is found
      if (minWeightVertex == endVertex) {
        break;
      }
      // if vertex has distance +infinity, then we have explored all accessible paths;
      // the remaining vertices in queue are *not* reachable from startVertex
      if (this.distances.get(minWeightVertex) == null) {
        break;
      }
      // follow outgoing edges
      for (Edge<Integer> outEdge : graph.getOutEdges(minWeightVertex)) {
        // check if triangle equation holds
        int newDist = this.distances.get(minWeightVertex) + outEdge.getInfo();
        int target = outEdge.getEndVertex();
        if (this.distances.get(target) == null || newDist < this.distances.get(target)) {
          // update weight and predecessor
          this.distances.put(target, newDist);
          predecessor.put(target, outEdge);
          // re-insert vertex to trigger queue update
          queue.remove(target);
          queue.offer(target);
        }
      }
    }

    // create edge list with shortest path
    List<Edge<Integer>> result = null;
    // is there a connection from start to end: distance end vertex < infinity?
    if (this.distances.get(endVertex) != null) {
      result = new LinkedList<>();
      Edge<Integer> curr = predecessor.get(endVertex);
      do {
        result.add(0, curr);
        curr = predecessor.get(curr.getStartVertex());
      } while (curr != null);
    }

    return result;
  }

  /** Comparator for directed graph vertices based on their distance. */
  static class DistComp implements Comparator<Integer> {

    // the distance for each vertex
    private Map<Integer, Integer> distances;

    /**
     * Create a new comparator operating on the given distances.
     *
     * @param distances a map with the distances
     */
    DistComp(Map<Integer, Integer> distances) {
      this.distances = distances;
    }

    @Override
    public int compare(Integer vertex1, Integer vertex2) {
      Integer dist1 = this.distances.get(vertex1);
      Integer dist2 = this.distances.get(vertex2);
      if (dist1 == null) {
        return (dist2 == null) ? 0 : 1;
      }
      if (dist2 == null) {
        return -1;
      }
      return Integer.compare(dist1, dist2);
    }
  }
}
