package de.unisaar.jfap.ex8;

import de.unisaar.jfap.ex7.DirectedGraph;
import de.unisaar.jfap.ex7.Edge;
import de.unisaar.jfap.ex7.visitors.DfsGraphVisitor;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.Stack;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Implement the {@link DfsGraphVisitor} interface for a visitor identifying the SCCs of a graph
 * using depth-first search.
 *
 * @author Bernd Kiefer, DFKI
 * @author Jörg Steffen, DFKI
 * @param <E> the type of information assigned to the edges of the traversed graph
 * @param <V> the type of information assigned to the vertices of the traversed graph
 */
public class TarjanVisitor<E, V> implements DfsGraphVisitor<E, V> {

  private static final Logger logger = LoggerFactory.getLogger(TarjanVisitor.class);

  // the time counter that is increased during DFS
  private int time;

  // the discovery time for each vertex
  private Map<Integer, Integer> discovery;

  // map each vertex to the lowest discovery time of a vertex reachable from it;
  // also use this to track if a vertex is still in the stack, i.e. is mapped to non-null
  private Map<Integer, Integer> low;

  // the vertices currently discovered but not assigned to a SCC yet
  private Stack<Integer> vertexStack;

  // the SCCs after the visitor has finished traversing the graph
  private Set<Set<Integer>> components;

  /** Create a new Tarjan visitor. */
  public TarjanVisitor() {
    this.discovery = new HashMap<>();
    this.low = new HashMap<>();
    this.vertexStack = new Stack<>();
    this.components = new HashSet<>();
    this.time = 0;
  }

  /**
   * Return the strongly connected components after the visitor has traversed the graph.
   *
   * @return set of SCCs represented as lists of integers
   */
  public Set<Set<Integer>> getSccs() {
    return this.components;
  }

  @Override
  public void startVertex(int v, DirectedGraph<E, V> g) {
    // nothing to do here
  }

  @Override
  public void discoverVertex(int v, DirectedGraph<E, V> g) {
    this.time++;
    // use low to detect if a vertex is still in the stack, see later
    this.low.put(v, this.time);
    this.discovery.put(v, this.time);
    this.vertexStack.push(v);
    logger.debug("discovered {} at {}", g.getVertexProperty(v), this.low.get(v));
  }

  @Override
  public void treeEdgeBefore(Edge<E> e, DirectedGraph<E, V> g) {
    // nothing to do here
  }

  @Override
  public void treeEdgeAfter(Edge<E> e, DirectedGraph<E, V> g) {
    // if the target is in another SCC and has been popped already,
    // its low value would be set to null again, so we have to check
    int endVertex = e.getEndVertex();
    if (this.low.containsKey(endVertex)) {
      int startVertex = e.getStartVertex();
      int startLow = this.low.get(startVertex);
      int endLow = this.low.get(endVertex);
      if (endLow < startLow) {
        this.low.put(startVertex, endLow);
        logger.debug(
            "lowering {} to {} after tree edge to {}",
            g.getVertexProperty(startVertex),
            endLow,
            g.getVertexProperty(endVertex));
      }
    } else {
      logger.debug(
          "ignore {} when discovered via tree edge from {}",
          g.getVertexProperty(endVertex),
          g.getVertexProperty(e.getStartVertex()));
    }
  }

  @Override
  public void nonTreeEdge(Edge<E> e, DirectedGraph<E, V> g) {
    int endVertex = e.getEndVertex();
    if (this.low.containsKey(endVertex)) {
      int startVertex = e.getStartVertex();
      int startLow = this.low.get(startVertex);
      int endLow = this.low.get(endVertex);
      if (endLow < startLow) {
        this.low.put(startVertex, endLow);
        logger.debug(
            "lowering {} to {} after non-tree edge to {}",
            g.getVertexProperty(startVertex),
            endLow,
            g.getVertexProperty(endVertex));
      }
    } else {
      logger.debug(
          "ignore {} when discovered via non-tree edge from {}",
          g.getVertexProperty(endVertex),
          g.getVertexProperty(e.getStartVertex()));
    }
  }

  @Override
  public void finishVertex(int v, DirectedGraph<E, V> g) {
    // do the computation of the lowest reachable vertex now
    logger.debug("finished {} at {}", g.getVertexProperty(v), this.low.get(v));
    if (this.discovery.get(v) == this.low.get(v)) {
      logger.debug("found new SCC with root {}", g.getVertexProperty(v));
      // all vertices on top of v in the stack and v itself belong to the SCC
      Set<Integer> component = new HashSet<>();
      int nextVertex = -1;
      do {
        nextVertex = this.vertexStack.pop();
        component.add(nextVertex);
        // we have to remove the low entry for the vertex so that it
        // doesn't interfere with later DFS runs
        this.low.remove(nextVertex);
      } while (nextVertex != v);
      this.components.add(component);
    }
  }
}
